
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdlib.h>
#include <sstream>
#include <iomanip>
#include <cmath>

#include "gall/gallnav.h" 
 
using namespace std;


// constructor
// ----------
t_gallnav::t_gallnav() 
  : t_gdata(),
    _offset(0),
    _nepoch(t_gtime::GPS),
    _overwrite(false)
{
  gtrace("t_gallnav::constructor");
  id_type(  t_gdata::ALLNAV );
  id_group( t_gdata::GRP_EPHEM);
}


// destructor
// ----------
t_gallnav::~t_gallnav()
{
  gtrace("t_gallnav::destructor");

  _mapsat.clear();
}


// return position
// ----------
int t_gallnav::pos( string sat, const t_gtime& t, double  xyz[],
		                                  double  var[],
		                                  double  vel[] ) // [m]
{
  gtrace("t_gallnav::pos");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  shared_ptr<t_geph> tmp = _find( sat, t );

  if( tmp == 0 ){	
    for(int i = 0; i<3; i++){
	                     xyz[i] = 0.0;
	           if( var ) var[i] = 0.0;
	           if( vel ) vel[i] = 0.0;
     }
     _gmutex.unlock(); return -1;
  }     
  int irc = tmp->pos( t, xyz, var, vel );
   
//  cout << "GALLNAV JSEM TU :: " << sat << " epoch: " << t.str_ymdhms()  << " toe: " << tmp->epoch().str_ymdhms() << endl; cout.flush();
   
  _gmutex.unlock(); return irc;
   
//  return find( sat, t )->pos( t, xyz, var, vel );
}


// return clock corrections
// ----------
int t_gallnav::clk( string sat, const t_gtime& t, double*  clk,
		                                  double*  var,
		                                  double* dclk ) // [s]
{
  gtrace("t_gallnav::clk");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
  shared_ptr<t_geph> tmp = _find( sat, t );
  if( tmp == 0 ){
              *clk  = 0.0;
    if( var ) *var  = 0.0;
    if(dclk ) *dclk = 0.0;
    _gmutex.unlock(); return -1;
  }

  int irc = tmp->clk( t, clk, var, dclk );
  _gmutex.unlock(); return irc;
   
//  return this->find( sat, t )->clk( t, clk, var, dclk );
}

// print function
// -------------------
void t_gallnav::print(string sat, const t_gtime& t)
{
  gtrace("t_gallnav::print");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
  shared_ptr<t_geph> tmp = _find( sat, t );
  tmp->print(); 
   
  _gmutex.unlock(); return;
}


// return list of available satellites
// ----------
vector<string> t_gallnav::satellites()
{
  gtrace("t_gallnav::satellites");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  vector<string> all_sat;
  map<string,t_map_ref>::const_iterator itPRN = _mapsat.begin();
   
  while( itPRN != _mapsat.end() ){
    all_sat.push_back(itPRN->first);
    itPRN++;
  }
  _gmutex.unlock(); return all_sat;
}


// return first position for satellite
// ----------
t_gtime t_gallnav::beg_gnav( string prn )
{
 gtrace("t_gallnav::beg_gnav");
   
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  t_gtime tmp = t_gtime::last_time();
  if( _mapsat.find(prn) != _mapsat.end() ) tmp = _mapsat[prn].begin()->first;

  _gmutex.unlock(); return tmp;
}


// return last position for satellite
// ----------
t_gtime t_gallnav::end_gnav( string prn )
{
 gtrace("t_gallnav::end_gnav");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  t_gtime tmp = t_gtime::first_time();
  if( _mapsat.find(prn) != _mapsat.end() ) tmp = _mapsat[prn].rbegin()->first;

  _gmutex.unlock(); return tmp;
}


// add navigation message
// ----------
int t_gallnav::add( shared_ptr<t_gnav> nav )
{   
 gtrace("t_gallnav::add");

//  if( nav->chk() < 0 ) return 1;
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  t_gtime ep(nav->epoch());
  string sat = nav->sat();
   
  if( _overwrite || _mapsat[sat].find(ep) == _mapsat[sat].end() ){

    if( _log && _log->verb() >= 2 ){
      ostringstream lg;
      lg << "add sat [" << nav->str_type()
         << "]: " << sat
         << " " << ep.str("%Y-%m-%d %H:%M:%S")
         << " iod: " << fixed << setw(3) <<  nav->iod();
      _log->comment(2,"gallnav",lg.str());
    }

    _mapsat[sat][ep] = nav;
  }
 
  _gmutex.unlock(); return 0;
}


// return number of epochs
// ----------
unsigned int t_gallnav::nepochs( const string& prn )
{
 gtrace("t_gallnav::nepochs");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  unsigned int tmp = 0;
  if( _mapsat.find(prn) != _mapsat.end() ) tmp = _mapsat[prn].size();
   
  _gmutex.unlock(); return tmp;
}


// list of epochs
// ----------
vector<t_gtime> t_gallnav::vec_epo( string prn )
{
 gtrace("t_gallnav::list_epo");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  vector<t_gtime> v_epo;

  t_map_sat::iterator itFIRST = _mapsat.begin();
  t_map_sat::iterator itLAST  = _mapsat.end();

  if( !prn.empty() ) itLAST++ = itFIRST = _mapsat.find( prn );

#ifdef DEBUG   
  if( itFIRST == itLAST          ) cout << "FIRST = LAST  \n";
  if( itFIRST == _mapsat.begin() ) cout << "FIRST = begin \n";
  if( itLAST  == _mapsat.end()   ) cout << "FIRST = end   \n";
#endif

  for( t_map_sat::iterator itSAT = itFIRST; itSAT != itLAST; ++itSAT )
    for( t_map_ref::iterator itEPO = itSAT->second.begin(); itEPO != itSAT->second.end(); ++itEPO )
      v_epo.push_back( itEPO->first );
   
  _gmutex.unlock(); return v_epo;
}

// list of nav messages
// ----------
vector<shared_ptr<t_geph>> t_gallnav::vec_nav( string prn )
{
 gtrace("t_gallnav::list_nav");

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  vector<shared_ptr<t_geph>> v_nav;

  t_map_sat::iterator itFIRST = _mapsat.begin();
  t_map_sat::iterator itLAST  = _mapsat.end();

  if( !prn.empty() ) itLAST++ = itFIRST = _mapsat.find( prn );

  for( t_map_sat::iterator itSAT = itFIRST; itSAT != itLAST; ++itSAT )
    for( t_map_ref::iterator itEPO = itSAT->second.begin(); itEPO != itSAT->second.end(); ++itEPO )
      v_nav.push_back( itEPO->second );

  _gmutex.unlock(); return v_nav;
}
   

// clean function
// ----------
void t_gallnav::clean_outer( const t_gtime& beg, const t_gtime& end )
{
 gtrace("t_gallnav::clean_outer");

  if( end < beg ) return;

#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

#ifdef DEBUG   
  if( _log ) _log->comment(1,"gallnav","rtcm clean request: "
			              + beg.str("%Y-%m-%d %H:%M:%S[%T] - ")
                                      + end.str("%Y-%m-%d %H:%M:%S[%T]"));
#endif

  // loop over all satellites
  map<string,t_map_ref>::const_iterator itPRN = _mapsat.begin();
  while( itPRN != _mapsat.end() ){
    string prn = itPRN->first;

    // find and CLEAN all data (epochs) out of the specified period !
    t_map_ref::iterator it;
    t_map_ref::iterator itFirst = _mapsat[prn].begin();
    t_map_ref::iterator itLast  = _mapsat[prn].end();
    t_map_ref::iterator itBeg   = _mapsat[prn].lower_bound(beg);  // greater only   old: // greater|equal
    t_map_ref::iterator itEnd   = _mapsat[prn].upper_bound(end);  // greater only!
     
//  cout << "distance = " << distance(itBeg,itEnd) << " prn " << prn << endl;

    // remove before BEGIN request
    if( itBeg != itFirst ){
       
      // begin is last
      if( itBeg == itLast ){
        itBeg--;
        if( _log && _log->verb() >= 2 ) 
	    _log->comment(2,"gallnav","gnav removed before: " + itFirst->first.str("%Y-%m-%d %H:%M:%S[%T] ") + prn);
        it = itFirst;
	while( it != itLast )

	  if( (it->second).use_count() == 0 ){ _mapsat[prn].erase(it++);
	  }else{ it++; }

      // begin is not last
      }else{
        if( _log && _log->verb() >=2 )
	    _log->comment(2,"gallnav","gnav removed before: " + itBeg->first.str("%Y-%m-%d %H:%M:%S[%T] ") + prn);
        it = itFirst;
	while( it != itBeg )
	  if( (it->second).use_count() == 0 ){ _mapsat[prn].erase(it++);
	  }else{ it++; }
	
      }
    }

    // remove after END request
    if( itEnd != itLast ){
        if( _log && _log->verb() >= 2 ) 
	    _log->comment(2,"gallnav","gnav removed after : " + itEnd->first.str("%Y-%m-%d %H:%M:%S[%T] ") + prn );
        it = itEnd;
	while( it != itLast )
	  if( (it->second).use_count() == 0 ){ _mapsat[prn].erase(it++);
	  }else{ it++; }

    }
    itPRN++;
  }
#ifdef BMUTEX   
  lock.unlock();
#endif
  _gmutex.unlock();
}


// return list of available satellites
// ----------
shared_ptr<t_geph> t_gallnav::_find( string sat, const t_gtime& t )
{
  gtrace("t_gallnav::_find sat/time");

  if( _mapsat.find(sat) == _mapsat.end() ) return make_shared<t_geph>();
  
  t_map_ref::iterator it = _mapsat[sat].lower_bound(t);  // greater|equal

  if (sat.compare(0,1,"R") != 0){
    if( it == _mapsat[sat].end() ){
      if( _log && _log->verb() >= 3 ){
        string lg("gnav element not found: [" + sat + t.str("] %Y-%m-%d %H:%M:%S "));
        _log->comment(3,"gallnav",lg);
      }       
      return make_shared<t_geph>();
    }
  }else{
     t_gtime tt = t - MAX_GLO_TIMEDIFF;  // time span of Glonass ephemerides is 15 min
     t_map_ref::iterator itt = _mapsat[sat].lower_bound(tt);  // greater|equal
     if( itt == _mapsat[sat].end() ) return  make_shared<t_geph>(); 
     double dt1 = itt->first - t;
     double dt2 =  it->first - t;
     if ( fabs(dt1) < fabs(dt2) ) it = itt;
  }
   
  return it->second;
}

// return list of available satellites
// ----------
shared_ptr<t_geph> t_gallnav::_find( string sat, int iod )
{
  gtrace("t_gallnav::_find sat/iod");

  if( _mapsat.find(sat) == _mapsat.end() ) return make_shared<t_geph>();

  t_map_ref::const_iterator it = _mapsat[sat].begin();
  while( it != _mapsat[sat].end() ){
    shared_ptr<t_gnav> pt_nav;
    if( ( pt_nav = dynamic_pointer_cast<t_gnav>(it->second) ) != NULL ){
//       cout << "nav->epoch() " << pt_nav->epoch().str_ymdhms() << " nav->iod() = " << pt_nav->iod()  << " rtcm iod: " << iod << endl;
      if( pt_nav->iod() == iod ) {
	 break; // found
      }
    }
    it++;
  }

  // not found !
  if( it == _mapsat[sat].end() ){
    if( _log && _log->verb() >= 3 ){
      ostringstream lg;
      lg << "warning: gnav element not found "
	 << "[sat: " << sat
	 << " iod: " << iod << "]";
      _log->comment(3,"gallnav",lg.str());
    }
    return make_shared<t_geph>();
  }

  return it->second;
}

