
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <cstring>
#include <memory>

#include "gdata/gnav.h"
#include "gdata/geph.h"
#include "gall/gallobj.h" 
#include "gall/gallnav.h"
#include "gutils/gtime.h"
#include "gcoders/rinexn.h"
#include "gutils/gtypeconv.h"

using namespace std;


/* ----------
 * constructor
 */
t_rinexn::t_rinexn( string version, int sz )
  : t_gcoder( version, sz )
{
  _gnsssys = 'G';
  _position = 0;
}


/* ----------
 * NAV-RINEX header
 */
int t_rinexn::decode_head( char* buff, int sz, vector<string>& errmsg)
{ 

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _mutex.lock();

  if( t_gcoder::_add2buffer( buff, sz) == 0 ){ _mutex.unlock(); return 0; };  

  string line;
  int consume = 0;
  int tmpsize = 0;
  while( ( tmpsize = t_gcoder::_getline( line )) >= 0 ){

    consume += tmpsize;

    if( line.find("RINEX VERSION",60) != string::npos ){          // first line
      switch ( line[20] ){
	case 'N': _gnsssys = 'G'; break; // Navigation data - according to Rinex specification
	case 'G': _gnsssys = 'R'; break; // GLONASS NAVIGATION - occures sometimes in brdc
        default : { string lg("warning: not rinex navigation data file");
                    if( _log )  _log->comment(0,"rinexn", lg);
                    else  cerr << lg;
                    _mutex.unlock(); return -1;
        }
      }
       
      switch ( line[40] ){
	case 'G': _gnsssys = 'G'; break; // GPS
	case 'R': _gnsssys = 'R'; break; // GLONASS
	case 'E': _gnsssys = 'E'; break; // GALILEO
	case 'J': _gnsssys = 'J'; break; // QZSS
	case 'S': _gnsssys = 'S'; break; // SBAS
	case 'M': _gnsssys = 'M'; break; // MIXED
        case ' ': if( line[20] == 'N' ) _gnsssys = 'G';
	          if( line[20] == 'G' ) _gnsssys = 'R';
	          if( _log ) 
	              _log->comment(0,"rinexn","warning - RINEXN system not defined, used "+t_gsys::char2str(_gnsssys));
	          break;
        default : { string lg("warning: undefined satellite system");
                    if( _log )  _log->comment(0,"rinexn", lg); 
	            else  cerr << lg;
                    // _mutex.unlock(); return -1;
        }
      }
       
      switch ( line[20] ){
	case 'N': break;  // Navigation data - according to Rinex specification
	case 'G': break;  // GLONASS NAVIGATION - occures sometimes in brdc	 
        default : { string lg("warning: not rinex navigation data file");
                    if( _log )  _log->comment(0,"rinexn", lg); 
	            else  cerr << lg;
//                    _mutex.unlock(); return -1;
        }
      }
       
//      if( _gnsssys == ' ' ){
//        if( _log ) _log->comment(0,"rinexn","warning: RINEX system not defined, GPS set as default" );
//        _gnsssys = 'G';
//      }

      _version = trim(line.substr(0,9));

      if( _log and substitute(_version, " ", "") > 0 ){
	_log->comment(2,"rinexn", "reading VER: " + _version + " SYS: " + string(1,_gnsssys) );
      }

    }else if( line.find("DELTA-UTC",60) != string::npos ){
#ifdef DEBUG
      double a0 = strSci2dbl(line.substr( 3,  19 ));
      double a1 = strSci2dbl(line.substr( 22, 19 ));
      int    w  = static_cast<int>(strSci2dbl(line.substr( 42,  8 )));
      int    t  = static_cast<int>(strSci2dbl(line.substr( 53,  6 )));
      if( _log ) _log->comment(2,"rinexn","reading DELTA-UTC");
      cout << "--- set DELTA-UTC: " << a0 << " : " << a1 << " : " << t << " : " << w << endl;
#endif

    }else if( line.find("PGM / RUN BY / DATE",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexn","reading PGM / RUN BY / DATE");
       
    }else if( line.find("IONOSPHERIC CORR",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexn","reading IONOSPHERIC CORR");
       
    }else if( line.find("TIME SYSTEM CORR",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexn","reading TIME SYSTEM CORR");

    }else if( line.find("CORR TO SYSTEM",60) != string::npos ){      // <= RINEX v2.10
      if( _log ) _log->comment(2,"rinexn","reading CORR TO SYSTEM");

    }else if( line.find("D-UTC A0,A1,T,W,S,U",60) != string::npos ){ // == RINEX v2.11
      if( _log ) _log->comment(2,"rinexn","reading D-UTC A0,A1,T,W,S,U");

    }else if( line.find("ION ALPHA",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexn","reading ION_ALPHA");

    }else if( line.find("ION BETA",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexn","reading ION_BETA");
       
    }else if( line.find("LEAP SECONDS",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexn","reading LEAP SECONDS");
       
    }else if( line.find("COMMENT",60) != string::npos ){
      if( _log ) _log->comment(2,"rinexn","reading COMMENT");
       
    }else if( line.find("END OF HEADER",60) != string::npos ){

      if( _log ) _log->comment(2,"rinexn","reading END OF HEADER ");
      t_gcoder::_consume(tmpsize);
      _mutex.unlock(); return -1;
    }
    t_gcoder::_consume(tmpsize);
  }

  _mutex.unlock(); return consume;
}

   
/* ----------
 * NAV-RINEX body
 */
int t_rinexn::decode_data(char* buff, int sz, int& cnt, vector<string>& errmsg)
{

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _mutex.lock();
   
  if( t_gcoder::_add2buffer(buff, sz) == 0 ){ _mutex.unlock(); return 0; };

#ifdef DEBUG
  cout << " BUFFER : \n" << _buffer << "\n size = " << sz  << " END OF BUFFER \n\n"; cout.flush();
#endif
  
  t_gtime epoch;
  string prn;
  int b, e, s, l, i;
  int maxrec = MAX_RINEXN_REC;  // implicite
  string timstr;
  t_gnavdata data = { 0 };

  // RINEX v2.xx   
  if( _version[0] == '2' ){        b = 0; e = 22; l = 19; s = 3; // timstr = "%2s %2d %02d %02d %02d %02d %02d";
  // RINEX v3.xx
  }else if( _version[0] == '3' ){  b = 0; e = 23; l = 19; s = 4; // timstr = "%3s %4d %02d %02d %02d %02d %02d";
  // RINEX ???
  }else{                           b = 0; e = 23; l = 19; s = 4; // timstr = "%3s %4d %02d %02d %02d %02d %02d";
  }

  string line;
  int consume = 0;
  int tmpsize = 0;
  int recsize = 0;

  while( ( tmpsize = t_gcoder::_getline( line, 0 ) ) >= 0 ){

#ifdef DEBUG
    cout << "0: " << line;
#endif

    consume += tmpsize;
    recsize += tmpsize;
    string epostr = line.substr(b,e);

    istringstream istr(line.substr(b,e) );
    istr.clear();

    prn = "";
    int yr,mn,dd,hr,mi,sc,svn = 0;
    double sec;

    if( _version[0] == '2' ){ istr >> svn >> yr >> mn >> dd >> hr >> mi >> sc;  sec = static_cast<double>(sc);
    }else{                    istr >> prn >> yr >> mn >> dd >> hr >> mi >> sec; }

    // check success
    if( ! istr.fail() ){

      shared_ptr<t_gnav> geph = make_shared<t_gnav>();
       
      // !!! SAT musi byt preveden na PRN (pro RINEX ver < 3) --> neni implementovano!
      if(       prn[0] == 'G' ){   maxrec = MAX_RINEXN_REC_GPS;  geph = make_shared<t_gnavgps>();  geph->glog(_log); epoch.tsys(t_gtime::GPS);
      }else if( prn[0] == 'R' ){   maxrec = MAX_RINEXN_REC_GLO;  geph = make_shared<t_gnavglo>();  geph->glog(_log); epoch.tsys(t_gtime::UTC);   // GLO  is equal to UTC time
      }else if( prn[0] == 'E' ){   maxrec = MAX_RINEXN_REC_GAL;  geph = make_shared<t_gnavgal>();  geph->glog(_log); epoch.tsys(t_gtime::GPS);   // GST  is equal to GPS time
      }else if( prn[0] == 'J' ){   maxrec = MAX_RINEXN_REC_QZSS; geph = make_shared<t_gnavqzss>(); geph->glog(_log); epoch.tsys(t_gtime::GPS);   // QZSS is equal to GPS time
      }else if( prn[0] == 'S' ){   maxrec = MAX_RINEXN_REC_SBAS; geph = make_shared<t_gnavsbas>(); geph->glog(_log); epoch.tsys(t_gtime::GPS);   // SBAS is equal to GPS time
      }else if( prn[0] == 'C' ){   maxrec = MAX_RINEXN_REC_BDS;  geph = make_shared<t_gnavbds>();  geph->glog(_log); epoch.tsys(t_gtime::GPS);   // BDT
      }else{

	 ostringstream tmpprn;
	 tmpprn << setw(1) << _gnsssys << setfill('0') << prn << setw(2) << svn ;
	 prn = tmpprn.str();

	 if(       _gnsssys == 'G' ){ maxrec = MAX_RINEXN_REC_GPS;  geph = make_shared<t_gnavgps>();  geph->glog(_log); epoch.tsys(t_gtime::GPS); 
         }else if( _gnsssys == 'R' ){ maxrec = MAX_RINEXN_REC_GLO;  geph = make_shared<t_gnavglo>();  geph->glog(_log); epoch.tsys(t_gtime::UTC);
         }else if( _gnsssys == 'E' ){ maxrec = MAX_RINEXN_REC_GAL;  geph = make_shared<t_gnavgal>();  geph->glog(_log); epoch.tsys(t_gtime::GPS);
         }else if( _gnsssys == 'J' ){ maxrec = MAX_RINEXN_REC_QZSS; geph = make_shared<t_gnavqzss>(); geph->glog(_log); epoch.tsys(t_gtime::GPS);
         }else if( _gnsssys == 'S' ){ maxrec = MAX_RINEXN_REC_SBAS; geph = make_shared<t_gnavsbas>(); geph->glog(_log); epoch.tsys(t_gtime::GPS);
         }else if( _gnsssys == 'C' ){ maxrec = MAX_RINEXN_REC_BDS;  geph = make_shared<t_gnavbds>();  geph->glog(_log); epoch.tsys(t_gtime::GPS);	    
	 }else{
	   cerr << "warning [rinexn]: undefined gnss system !\n";
//	   break;
	 }
      }
      epoch.from_ymdhms(yr, mn, dd, hr, mi, sec);

#ifdef DEBUG
      cout << "  satell = " << prn 
	   << "  maxrec = " << maxrec 
	   << epoch.str_ymdhms("  ep = ")
	   << endl; cout.flush();
#endif

      data[0] = strSci2dbl(line.substr( 19+s, l ));
      data[1] = strSci2dbl(line.substr( 38+s, l ));
      data[2] = strSci2dbl(line.substr( 57+s, l ));
       
      i = 2;
      while( i < MAX_RINEXN_REC ){

	if( ( tmpsize = t_gcoder::_getline( line, recsize ) ) < 0 ){
//	  cout << "neni cely zaznam ! " << tmpsize << " " << recsize << " " << consume << "\n"; cout.flush();
	  break;
        }
        consume += tmpsize;
        recsize += tmpsize;
        if( ++i < maxrec ){ data[i] = strSci2dbl(line.substr(    s, l )); } //  cout << "1 => " << data[i] << "\n"; }
        if( ++i < maxrec ){ data[i] = strSci2dbl(line.substr( 19+s, l )); } //  cout << "2 => " << data[i] << "\n"; }
        if( ++i < maxrec ){ data[i] = strSci2dbl(line.substr( 38+s, l )); } //  cout << "3 => " << data[i] << "\n"; }
        if( ++i < maxrec ){ data[i] = strSci2dbl(line.substr( 57+s, l )); } //  cout << "4 => " << data[i] << "\n"; }
#ifdef DEBUG
      cout << "prn = " << prn << " " << prn[0] << "  maxrec = " << maxrec 
	   << " " << i << epoch.str_ymdhms("  ep = ")
           << " " << data[i-3] << " " << data[i-2] << " " << data[i-1] << " " << data[i-0] << endl;
#endif
        // is record complete and filter-out GNSS systems
        if( geph && i+1 >= maxrec )
	{
          t_gcoder::_consume(recsize);
          recsize = 0;

	  // skip non-valid GNSS
	  if( _sys.size() != 0 && 
	      _sys.find( t_gsys::char2str( prn[0] )) == _sys.end() ) break;

          geph->data2nav( prn, epoch, data );

	  string msg;
	  if( geph->chk(msg) < 0 ){
	    if( _log ) _log->comment(1,"rinexn","skipped sat " + prn + " " + msg);
	    break; 
	  }

          if( _log ) _log->comment(2,"rinexn","add sat nav " + prn + epoch.str_ymdhms(" ") );

          // fill data
          map<string,t_gdata*>::iterator it = _data.begin();
	  while( it != _data.end() ){

            if (it->second->id_type()  == t_gdata::ALLNAV ||
	        it->second->id_group() == t_gdata::GRP_EPHEM){

	       ((t_gallnav*)it->second)->add( geph );

            }else if (it->second->id_type() == t_gdata::ALLOBJ){
	       if (geph->id_type() == t_gdata::EPHGLO){
	         t_gallobj* all_obj = (t_gallobj*)it->second;
	         shared_ptr<t_gobj>  one_obj = all_obj->obj(geph->sat());
	         shared_ptr<t_gobj>  objnull;
	         int ch = dynamic_pointer_cast<t_gnavglo>(geph)->channel();	         
	         if (one_obj != objnull) {
		    one_obj->channel(ch);
		 }		  
	       }
            }
  	    it++;
          }
          cnt++;
          break;
	}	 
      }
      if( recsize != 0 ) break; // break not complete record

   }else{

     // remove this data
     t_gcoder::_consume(tmpsize);
     recsize = 0;
     consume = 0;

#ifdef DEBUG
     cout << "\nremoving from buffer [" << tmpsize << "/" << line << "]\n\n";
#endif
       
   }

   if( recsize != 0 ) break; // break the initialization loop if read not finished correctly
 }
 _mutex.unlock(); return consume;
}


/* -------- *
 * ENCODER
 * -------- */

/* ----------
 * NAV-RINEX header
 */
int t_rinexn::encode_head( char* buff, int sz, vector<string>& errmsg)
{ 

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _mutex.lock();

  if( _position == 0 ){

    t_gtime tt(t_gtime::UTC);

    ss << "     3.02           N: GNSS NAV DATA    M (MIXED)           RINEX VERSION / TYPE" << endl
    // << "G-Nut/Anubis        GOP/RIGTC           20120618 000000 UTC PGM / RUN BY / DATE"  << endl
       << "G-Nut/Anubis        GOP/RIGTC           "
       << tt.str("%Y%m%d %H%M%S UTC")                             << " PGM / RUN BY / DATE"  << endl
       << "                                                            END OF HEADER"        << endl;
  }

  ss.seekg(0, ss.end);
  long len = ss.tellg();
  len = len - _position;
  ss.seekg(_position, ios_base::beg);

  int size = ( len < sz ) ? len : sz;

  ss.clear();
  ss.read(buff,size);

   
  if( ss.fail() ){ cout << "HEAD: any problem ?\n";

  }else if( ss.gcount() == 0 ){
//  cout << "HEAD: finished " << sz << " " << len << " " << _position << " " << ss.gcount() << endl;
    _position = 0;
    ss.str(""); 
    ss.clear();
  }else{
//  cout << "HEAD: read     " << sz << " " << len << " " << _position << " "  << ss.gcount()<< endl;
    _position += size; // ss.gcount();
  }
   
  _mutex.unlock(); return size;
}


/* ----------
 * NAV-RINEX data
 */
int t_rinexn::encode_data( char* buff, int sz, int& cnt, vector<string>& errmsg)
{ 

#ifdef BMUTEX
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _mutex.lock();

  // fill data
  if( _position == 0 ){

    map<string,t_gdata*>::iterator it = _data.begin();
    for( it = _data.begin(); it != _data.end(); ++it ){

      if( it->second->id_type()  == t_gdata::ALLNAV ||
          it->second->id_group() == t_gdata::GRP_EPHEM ){

        vector<shared_ptr<t_geph>> vec_nav = dynamic_cast<t_gallnav*>(it->second)->vec_nav();
        vector<shared_ptr<t_geph>>::iterator itNAV;
        for( itNAV = vec_nav.begin(); itNAV != vec_nav.end(); ++itNAV )
        {
          shared_ptr<t_gnav> gnav = dynamic_pointer_cast<t_gnav>(*itNAV);

//          cout << gnav->sat() << " " << gnav->epoch().str_ymdhms(" ") << endl;
   
          t_gnavdata data;
	  if( gnav->nav2data( data ) == 0 ){

            ss << gnav->sat() + " " + gnav->epoch().str("%Y %m %d %H %M %S")
	       << fixed      // << setprecision(1)  << setw(4) << gnav->epoch().sec() + gnav->epoch().dsec()
	       << scientific << setprecision(12); // setfill('0')
	    for( int i=0; i<gnav->rec(); ++i ){
	      if( (i-3)%4 == 0 ) ss << endl << setw(4) << ""; // space before first column first
              ss << setw(19) << data[i];
	    }
	    ss << endl;
	  }
        }
      }
    }
  }

  ss.seekg(0, ss.end);
  long len = ss.tellg();
  len = len - _position;
  ss.seekg(_position, ios_base::beg);

  int size = ( len < sz ) ? len : sz;

  ss.clear();
  ss.read(buff,size);
   
  if( ss.fail() ){ cout << "DATA: any problem ?\n";

  }else if( ss.gcount() == 0 ){
//cout << "DATA: finished " << sz << " " << len << " " << _position << " " << ss.gcount() << endl;
    _position = 0;
    ss.str(""); 
    ss.clear();

  }else{
//cout << "DATA: read     " << sz << " " << len << " " << _position << " "  << ss.gcount()<< endl;
    _position += size; // ss.gcount();
  }

  _mutex.unlock(); return size;
}
