/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include <sstream>

#include "gdata/gobsgnss.h"
#include "gutils/gconst.h"
#include "gutils/gtypeconv.h"

using namespace std;

// constructor
// -----------
t_gobsgnss::t_gobsgnss()
  : t_gdata(),
     _channel(DEF_CHANNEL)
{ 
  id_type(t_gdata::OBSGNSS);
  id_group(t_gdata::GRP_OBSERV);
}

t_gobsgnss::t_gobsgnss(const string& site, const string& sat, const t_gtime& t)
  : t_gdata(),
     _staid(site),
     _satid(sat),
     _epoch(t),
     _channel(DEF_CHANNEL)
{ 
  id_type(t_gdata::OBSGNSS);
  id_group(t_gdata::GRP_OBSERV);
}


// destructor
// -----------
t_gobsgnss::~t_gobsgnss()
{
#ifdef DEBUG
  cout << "OBS-destruct: " << site() << " " << sat() << " " 
       << epoch().str("  %Y-%m-%d %H:%M:%S[%T] ") << fixed << setprecision(3);
   
  vector<GOBS> v_obs = this->getobs();
  vector<GOBS>::iterator itOBS = v_obs.begin();
  for( ; itOBS != v_obs.end(); ++itOBS ) 
      cout << " " << t_gobs::gobs2str( *itOBS ) << ":" << this->getobs( *itOBS );
  cout  << endl; cout.flush();   
#endif
}


// add observation - special map of doubles
// -----------
void t_gobsgnss::addobs(const GOBS& obs, const double& d)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
  _gobs[obs] = d;
  _gmutex.unlock();
}


// add lost-of-lock indicators (LLI) - special map of integers
// -----------
void t_gobsgnss::addlli(const GOBS& obs, const int& i)
{  
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
  if( i != 0 ) _glli[obs] = i;
  _gmutex.unlock();
}


// get vector of observations types (GOBS)
// -----------
vector<GOBS> t_gobsgnss::obs()
{  
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  vector<GOBS> tmp;
  map<GOBS,double>::iterator it = _gobs.begin();  
  while( it != _gobs.end() ){
    tmp.push_back( it->first );
    it++;
  }
   
  _gmutex.unlock(); return tmp;
}

// add OBS - code [m], phase [whole cycles], dopler[cycles/sec], snr [DBHZ], ...  observations (double)
// -----------
double t_gobsgnss::getobs(const GOBS& obs)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  double tmp = NULL_GOBS;
  if( _gobs.find(obs) != _gobs.end() ) tmp = _gobs[obs];
   
  _gmutex.unlock(); return tmp;
}


// add OBS - code [m], phase [whole cycles], dopler[cycles/sec], snr [DBHZ], ...  observations (double)
// -----------
double t_gobsgnss::getobs(const string& obs)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  double tmp = NULL_GOBS;
  if( _gobs.find(t_gobs::str2gobs(obs)) != _gobs.end() ){ tmp = _gobs[t_gobs::str2gobs(obs)]; }

  _gmutex.unlock(); return tmp;
}


// add LLI - lost of lock indicator (int)
// -----------
int t_gobsgnss::getlli(const GOBS& obs)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  int tmp = 0;
  if( _glli.find(obs) != _glli.end() ) tmp = _glli[obs];
   
  _gmutex.unlock(); return tmp;
}


// add LLI - lost of lock indicator (int)
// -----------
int t_gobsgnss::getlli(const string& obs)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  int tmp = 0;
  if( _glli.find(t_gobs::str2gobs(obs)) != _glli.end() ) tmp = _glli[t_gobs::str2gobs(obs)];
   
  _gmutex.unlock(); return tmp;
}


// set chanel number for Glonass satellites
// -----------------
void t_gobsgnss::channel(int ch)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
  _channel = ch;
  _gmutex.unlock();
}


// get chanel number for Glonass satellites
// -----------------
int t_gobsgnss::channel() const
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  int tmp = _channel;
   
  _gmutex.unlock(); return tmp;
}

// return # of observations
size_t t_gobsgnss::size() const
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  size_t tmp = _gobs.size();
   
  _gmutex.unlock(); return tmp;
}

// return system-dependenty frequency multiplier factor
// ---------- 
double t_gobsgnss::factor( int i )
{
  switch( this->gsys() ){
     
   case GPS  : switch( i ){
                 case 1 : return G01_MLT;
                 case 2 : return G02_MLT;
                 case 5 : return G05_MLT;
                default : return 0.0; }
     
   case GLO  : switch( i ){
                 case 1 : return R01_MLT;
                 case 2 : return R02_MLT;
                default : return 0.0; }
     
   case GAL  : switch( i ){
                 case 1 : return G01_MLT;
                 case 5 : return E05_MLT;
                 case 6 : return E06_MLT;
                 case 7 : return E07_MLT;
                 case 8 : return E08_MLT;
                default : return 0.0; }
     
   case BDS  : switch( i ){
                 case 1 : return C01_MLT;
                 case 6 : return C06_MLT;
                 case 7 : return C07_MLT;
                default : return 0.0; }
     
   case SBAS : switch( i ){
                 case 1 : return J01_MLT;
                 case 2 : return J02_MLT;
                 case 5 : return J05_MLT;
                 case 6 : return J06_MLT;
                default : return 0.0; }
     
   case QZSS : switch( i ){
                 case 1 : return S01_MLT;
                 case 5 : return S05_MLT;
                default : return 0.0; }
     
   case GNSS : return 0.0;
  }

  return 0.0;
}

// // get system-specific frequency for band i
// ---------- 
double t_gobsgnss::frequency( int i )
{
  switch( this->gsys() ){
     
   case GPS  : switch( i ){
                 case 1 : return G01_F;
                 case 2 : return G02_F;
                 case 5 : return G05_F;
                default : return 0.0; }
     
   case GLO  : switch( i ){
                 case 1 : return R01_F(_channel); 
                 case 2 : return R02_F(_channel); 
                default : return 0.0; }
     
   case GAL  : switch( i ){
                 case 1 : return G01_F;
                 case 5 : return E05_F;
                 case 6 : return E06_F;
                 case 7 : return E07_F;
                 case 8 : return E08_F;
                default : return 0.0; }
     
   case BDS  : switch( i ){
                 case 1 : return C01_F;
                 case 6 : return C06_F;
                 case 7 : return C07_F;
                default : return 0.0; }
     
   case QZSS : switch( i ){
                 case 1 : return J01_F;
                 case 2 : return J02_F;
                 case 5 : return J05_F;
                 case 6 : return J06_F;
                default : return 0.0; }
     
   case SBAS : switch( i ){
                 case 1 : return S01_F;
                 case 5 : return S05_F;
                default : return 0.0; }
     
   case GNSS : return 0.0;
  }

  return 0.0;
}


// get system-specific wavelength for band i
// ---------- 
double t_gobsgnss::wavelength( int i )
{
  double frq = this->frequency( i );
  if( frq != 0.0 ) return CLIGHT/frq;
     
  return 0.0;
}

// get wavelength for iono-fee LC band i,j
// ---------- 
double t_gobsgnss::wavelength_L3( int b1, int b2 )
{
  double c1, c2; 
  ionofree_coef(b1, c1, b2, c2);   
   
  double f1 = this->frequency( b1 );
  double f2 = this->frequency( b2 );
   
  if ( !double_eq(f1, 0.0) && !double_eq(f2, 0.0) ){
     double lamb_L3 = c1 * CLIGHT/f1 + c2 * CLIGHT/f2;     
     return lamb_L3;
  }
   
   return 0.0;   
}

//  get GNSS system from satellite IDOC
// ----------
GSYS t_gobsgnss::gsys()
{
  return t_gsys::char2gsys( _satid[0] );
}


// get selected code (GOBS)
// ----------
GOBS t_gobsgnss::cod_id(const int& band) // , const GOBS& obs)
{    
  size_t sz = sizeof(code_choise[band]) / sizeof(GOBS);
  map<GOBS,double>::iterator it; // = _gobs.find( obs );

  GOBS tmp = X; // obs;

  // choices 
  for( size_t i=0; i<sz; ++i ){
    tmp = code_choise[band][i];
    
    it = _gobs.find( tmp );
    if( tmp == X || it == _gobs.end() ) continue;
       
    if( ! double_eq(it->second, NULL_GOBS ) ) return tmp;
//    { cout << " found  code ID: [" << band << "] " << t_gobs::gobs2str( tmp ) << endl; return tmp; }
  }

  return X;
}


// get selected phase (GOBS)
// ----------
GOBS t_gobsgnss::pha_id(const int& band) // , const GOBS& obs)
{    
  size_t sz = sizeof(phase_choise[band]) / sizeof(GOBS);
  map<GOBS,double>::iterator it; // = _gobs.find( obs );
 
  GOBS tmp = X; // obs;

  // choices
  for( size_t i=0; i<sz; ++i ){
    tmp = phase_choise[band][i];
       
    it = _gobs.find( tmp );
    if( tmp == X || it == _gobs.end() ) continue;

    if( ! double_eq( it->second, NULL_GOBS) ) return tmp;
//    { cout << " found phase ID: [" << band << "] " << t_gobs::gobs2str( tmp ) << endl; return tmp; }
  }

  return X;
}


// get selected code (GOBS)
// ----------
GOBS t_gobsgnss::code(const int& band, const GOBS& obs)
{    
  size_t sz = sizeof(code_choise[band]) / sizeof(GOBS);
  map<GOBS,double>::iterator it = _gobs.find( obs );

  // not preffered option !
  if( obs == X ){
    
    GOBS tmp = obs;

    // loop over choices 
    for( size_t i=0; i<sz; i++ ){
      tmp = code_choise[band][i];
    
      it = _gobs.find( tmp );
      if( tmp == X || it == _gobs.end() ) continue;
       
      if( ! double_eq(it->second, NULL_GOBS ) ) return tmp;
       // { cout << " found  code ID: [" << band << "] " << t_gobs::gobs2str( tmp ) << endl; return tmp; }
    }

  // preffered option (via obs settings)
  }else if( it != _gobs.end() && ! double_eq( it->second, NULL_GOBS) ) return obs;
    // { cout << " requested code ID: [" << band << "] " << t_gobs::gobs2str( obs ) << endl; return obs; }

  return X;
}


// get selected phase (GOBS)
// ----------
GOBS t_gobsgnss::phase(const int& band, const GOBS& obs)
{    
  size_t sz = sizeof(phase_choise[band]) / sizeof(GOBS);
  map<GOBS,double>::iterator it = _gobs.find( obs );
 
  // not preffered option !
  if( obs == X ){
    
    GOBS tmp = obs;

    // loop over choices 
    for( size_t i=0; i<sz; i++ ){
      tmp = phase_choise[band][i];
       
      it = _gobs.find( tmp );
      if( tmp == X || _gobs.find( tmp ) == _gobs.end() ) continue;

      if( ! double_eq( it->second, NULL_GOBS) ) return tmp;
      // { cout << " found phase ID: [" << band << "] " << t_gobs::gobs2str( tmp ) << endl; return tmp; }
    }

  // preffered option (via obs settings)
  }else if( it != _gobs.end() && ! double_eq( it->second, NULL_GOBS) ) return obs;
   // { cout << " requested phase ID: [" << band << "] " << t_gobs::gobs2str( obs ) << endl; return obs; }

  return X;
}


// code observations [m]
// ----------
double t_gobsgnss::obs_C(const int& band)
{   
//  boost::mutex::scoped_lock lock(_mutex);
  GOBS tmpobs = X;
  tmpobs = cod_id(band);
  
  if( tmpobs != X ){
    return _gobs[tmpobs];
  }

  return NULL_GOBS; // not found
}


// code observations [m]
// ----------
double t_gobsgnss::obs_C(const GOBS& gobs)
{   
//  boost::mutex::scoped_lock lock(_mutex);
   
  // Glonass has 255 channel number 
  if ( ! _valid_obsL() ) return 0.0; 
   
  map<GOBS,double>::iterator it = _gobs.find( gobs );
   
  if( gobs != X )
    return ( it->second );

  return NULL_GOBS; // not found
}


// phase observations [m]
// ----------
double t_gobsgnss::obs_L(const int& band)
{   
//  boost::mutex::scoped_lock lock(_mutex);

  GOBS tmpobs = X;
  tmpobs = pha_id(band);
  t_gsys gsys( this->gsys() );

#ifdef DEBUG
  cout << " obs_L("  << band << ")  tmpobs = " << t_gobs::gobs2str( tmpobs )
       << " gsys = " << gsys.str() << " prn = " << _satid << " wvlength = " << this->wavelength( band ) << " channel = " << _channel << endl;
#endif

  if( tmpobs != X )
    return ( _gobs[tmpobs] * this->wavelength(band) ); // transfer from whole cycles to meters!

  return NULL_GOBS; // not found
}


// phase observations [m]
// ----------
double t_gobsgnss::obs_L(const GOBS& gobs)
{   
//  boost::mutex::scoped_lock lock(_mutex);
  map<GOBS,double>::iterator it = _gobs.find( gobs );
  t_gsys gsys( this->gsys() );

  if( gobs != X )
  return ( it->second * this->wavelength(t_gobs::gobs2band(gobs)) ); // transfer from whole cycles to meters!

  return NULL_GOBS; // not found
}


// modify phase observations [m]
// ----------
int t_gobsgnss::mod_L(const double& dL, const GOBS& gobs, const int i)
{   
//  boost::mutex::scoped_lock lock(_mutex);

  map<GOBS,double>::iterator it;
  t_gsys gsys( this->gsys() );

  if( gobs == X ) // mo
  {	
    for(it = _gobs.begin(); it != _gobs.end(); ++it )
    {
      // phase only !!!!  
      string gobs_str = t_gobs::gobs2str(it->first);
      if (gobs_str.compare(0,1,"L") != 0) continue;
       
      int band = t_gobs::gobs2band(it->first);
      if (i == 1) it->second += dL/this->wavelength(band);  // transfer back to cycles
      else if (i == 0) it->second += dL; 
      else cout << "Observations not modified!" << endl;
//      cout << "! POZOR ! modifikuji: " << t_gobs::gobs2str(it->first) << " o " << dL << " band " << band << endl;
    }

  // modify requested only
  }else{
    it = _gobs.find(gobs);
    if( it == _gobs.end() ) return -1;
     
    int band = t_gobs::gobs2band(gobs);
    if (i == 1) it->second += dL/this->wavelength(band);  // transfer back to cycles
    else if (i == 0) it->second += dL; 
    else cout << "Observations not modified!" << endl;     
//    cout << "! POZOR ! modifikuji: " << t_gobs::gobs2str(gobs) << " o " << dL << " band " << band << endl;

  }

  return 0; // not found
}


// return value of general code linear combination
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::code_lc(const int& band1, const double& coef1,
	                   const int& band2, const double& coef2,
 	                   const int& band3, const double& coef3 )
{
  double C1 = NULL_GOBS;
  double C2 = NULL_GOBS;
  double C3 = NULL_GOBS;
   
  if( coef1 != 0.0 ){ C1 = obs_C( band1 ); if( double_eq(C1,NULL_GOBS) ) return NULL_GOBS; }
  if( coef2 != 0.0 ){ C2 = obs_C( band2 ); if( double_eq(C2,NULL_GOBS) ) return NULL_GOBS; }
  if( coef3 != 0.0 ){ C3 = obs_C( band3 ); if( double_eq(C3,NULL_GOBS) ) return NULL_GOBS; }

  double lc = ( coef1*C1 + coef2*C2 + coef3*C3 );

#ifdef DEBUG
    cout << _satid << " " << _staid
         << "  code_lc BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<         "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ") "
         << fixed << setprecision(3) << setw(16) << C1 << setw(16) << C2 << setw(16) << C3 << "  lc = " << setw(16) << lc <<  endl;
#endif
   
  return lc;
}


// return value of general carrier-phase linear combination
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::phase_lc(const int& band1, const double& coef1,
	                    const int& band2, const double& coef2,
	                    const int& band3, const double& coef3 )
{
  double L1 = NULL_GOBS;
  double L2 = NULL_GOBS;
  double L3 = NULL_GOBS;

  if( coef1 != 0.0 ){ L1 = obs_L( band1 ); if( double_eq(L1,NULL_GOBS) ) return NULL_GOBS; }
  if( coef2 != 0.0 ){ L2 = obs_L( band2 ); if( double_eq(L2,NULL_GOBS) ) return NULL_GOBS; }
  if( coef3 != 0.0 ){ L3 = obs_L( band3 ); if( double_eq(L3,NULL_GOBS) ) return NULL_GOBS; }

  double lc = ( coef1*L1 + coef2*L2 + coef3*L3 );

#ifdef DEBUG
    cout << _satid << " " << _staid
         << " phase_lc BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<         "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ") "
         << fixed << setprecision(3) << setw(16) << L1 << setw(16) << L2 << setw(16) << L3 << "  lc = " << setw(16) << lc << endl;
//         << " TEST :" <<  t_gobs::gobs2str(pha_id(band1))
//         << " "       <<  getobs( pha_id(band1) ) << endl
//         << " "       <<  getobs( pha_id(band1) ) << endl
//         << " obs_L:" <<  obs_L( band1 ) << endl;
#endif
  return lc;
}

// return value of carrier-phase linear combination frequency
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::frequency_lc(const int& band1, const double& coef1,
				const int& band2, const double& coef2,
				const int& band3, const double& coef3 )
{  
  double f1 = 0.0;
  double f2 = 0.0;
  double f3 = 0.0;

  if( coef1 != 0.0 ){ f1 = frequency( band1 ); if( double_eq(f1,NULL_GOBS) ) return NULL_GOBS; }
  if( coef2 != 0.0 ){ f2 = frequency( band2 ); if( double_eq(f2,NULL_GOBS) ) return NULL_GOBS; }
  if( coef3 != 0.0 ){ f3 = frequency( band3 ); if( double_eq(f3,NULL_GOBS) ) return NULL_GOBS; }

  double freq = ( coef1*f1 + coef2*f2 + coef3*f3 );

#ifdef DEBUG
    cout << _satid << " " << _staid
         << " frequency_lc BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<         "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ") "
         << fixed << setprecision(3) << setw(16) << f1 << setw(16) << f2 << setw(16) << f3 << "  freq = " << setw(16) << freq << endl;
#endif
  return freq;
}

// return value of carrier-phase linear combination ionosphere scale factor
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::isf_lc(const int& band1, const double& coef1,
		  	  const int& band2, const double& coef2,
			  const int& band3, const double& coef3 )
{  
  double f1 = 1.0;
  double f2 = 1.0;
  double f3 = 1.0;

  if( coef1 != 0.0 ){ f1 = frequency( band1 ); if( double_eq(f1,NULL_GOBS) ) return NULL_GOBS; }
  if( coef2 != 0.0 ){ f2 = frequency( band2 ); if( double_eq(f2,NULL_GOBS) ) return NULL_GOBS; }
  if( coef3 != 0.0 ){ f3 = frequency( band3 ); if( double_eq(f3,NULL_GOBS) ) return NULL_GOBS; }

  double denom = coef1*f1 + coef2*f2 + coef3*f3;
  double num   = f1*f1*(coef1/f1 + coef2/f2 + coef3/f3);
  double isf   = num/denom;
     
#ifdef DEBUG
    cout << _satid << " " << _staid
         << " isf_lc BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<         "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ") "
         << fixed << setprecision(3) << setw(16) << f1 << setw(16) << f2 << setw(16) << f3 << "  isf = " << setw(16) << isf << endl;
#endif
  return isf;
}

// return value of general code linear combination
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::code_lcf(const int& band1, const int& coef1,
	 	            const int& band2, const int& coef2,
		            const int& band3, const int& coef3 )
{
  double C1 = NULL_GOBS;
  double C2 = NULL_GOBS;
  double C3 = NULL_GOBS;
   
  if( coef1 != 0.0 ){ C1 = obs_C( band1 ); if( double_eq(C1,NULL_GOBS) ) return NULL_GOBS; }
  if( coef2 != 0.0 ){ C2 = obs_C( band2 ); if( double_eq(C2,NULL_GOBS) ) return NULL_GOBS; }
  if( coef3 != 0.0 ){ C3 = obs_C( band3 ); if( double_eq(C3,NULL_GOBS) ) return NULL_GOBS; }

  double f1 = this->frequency( band1 );
  double f2 = this->frequency( band2 );
  double f3 = this->frequency( band3 );

  double lc =   ( coef1*f1*C1 + coef2*f2*C2 + coef3*f3*C3 )
              / ( coef1*f1    + coef2*f2    + coef3*f3    );

#ifdef DEBUG
    cout << _satid << " " << _staid
         << "  code_lcf BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<          "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ")  "
         << fixed << setprecision(3) << setw(16) << C1 << setw(16) << C2 << setw(16) << C3 << "  lc = " << setw(16) << lc <<  endl;
#endif
   
  return lc;
}
   

// return value of general carrier-phase linear combination
// for 2 or 3 bands with given coefficients
// ----------
double t_gobsgnss::phase_lcf(const int& band1, const int& coef1,
		             const int& band2, const int& coef2,
		             const int& band3, const int& coef3 )
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();   
   
   if (gsys() == GLO && double_eq(_channel, DEF_CHANNEL)) {
      _gmutex.unlock(); 
      return 0.0;
   }   
   
  double L1 = NULL_GOBS;
  double L2 = NULL_GOBS;
  double L3 = NULL_GOBS;
  if( coef1 != 0.0 ){ L1 = obs_L( band1 ); if( double_eq(L1,NULL_GOBS) ) {_gmutex.unlock(); return NULL_GOBS;} }
  if( coef2 != 0.0 ){ L2 = obs_L( band2 ); if( double_eq(L2,NULL_GOBS) ) {_gmutex.unlock(); return NULL_GOBS;} }
  if( coef3 != 0.0 ){ L3 = obs_L( band3 ); if( double_eq(L3,NULL_GOBS) ) {_gmutex.unlock(); return NULL_GOBS;} }

  double f1 = this->frequency( band1 );
  double f2 = this->frequency( band2 );
  double f3 = this->frequency( band3 );

  double lc =   ( coef1*f1*L1 + coef2*f2*L2 + coef3*f3*L3 )
              / ( coef1*f1    + coef2*f2    + coef3*f3    );

#ifdef DEBUG
    cout << _satid << " " << _staid
         << " phase_lcf BAND(" << band1 << "," << band2 << "," << band3 << ") "
         <<          "  COEF(" << coef1 << "," << coef2 << "," << coef3 << ") "
         << fixed << setprecision(3) << setw(16) << L1 << setw(16) << L2 << setw(16) << L3 << "  lc = " << setw(16) << lc <<  endl;
#endif

   _gmutex.unlock();
   return lc;
}


// return coefficients (c1,c2) of the ionosphere-free linear combination
// ----------
int t_gobsgnss::ionofree_coef( const int& b1, double& c1, 
                               const int& b2, double& c2 )
{
  if( b1 == b2 ){ c1 = c2 = 0.0; return -1; }
  t_gsys gsys( this->gsys() );
   
  double fac1 = gsys.factor( b1 );
  double fac2 = gsys.factor( b2 );
   
  c1 =   fac1*fac1 / (fac1*fac1 - fac2*fac2);
  c2 = - fac2*fac2 / (fac1*fac1 - fac2*fac2);
   
#ifdef DEBUG
  cout << " coefs (D) = " << fixed << setprecision(3)
       << setw(15) << fac1
       << setw(15) << fac2
       << setw(15) << c1
       << setw(15) << c2
       << endl;
#endif

  return 0;
}


// return coefficients (c1,c2) of the geometry-free linear combination
// ----------
int t_gobsgnss::geomfree_coef(const int& b1, double& c1, 
                              const int& b2, double& c2)
{   
  if( b1 == b2 ){ c1 = c2 = 0.0; return -1; }
  c1 =  1.0;
  c2 = -1.0;

  return 0;
}


// return coefficients (c1,c2) of the narrow-lane linear combination
// ----------
int t_gobsgnss::narrlane_coef(const int& b1, double& c1,
                              const int& b2, double& c2)
{   
  if( b1 == b2 ){ c1 = c2 = 0.0; return -1; }
  t_gsys gsys( this->gsys() );

  double fac1 = gsys.factor( b1 );
  double fac2 = gsys.factor( b2 );
   
  c1 =   fac1 / (fac1 + fac2);
  c2 =   fac2 / (fac1 + fac2);

  return 0;
}


// return coefficients (c1,c2) of the wide-lane linear combination
// ----------
int t_gobsgnss::widelane_coef(const int& b1, double& c1, 
                              const int& b2, double& c2)
{   
  if( b1 == b2 ){ c1 = c2 = 0.0; return -1; }
  t_gsys gsys( this->gsys() );

  double fac1 = gsys.factor( b1 );
  double fac2 = gsys.factor( b2 );
   
  c1 =   fac1 / (fac1 - fac2);
  c2 = - fac2 / (fac1 - fac2);

  return 0;
}


// return coefficients (c1,c2) of the code multipath linear combination
// ----------
int t_gobsgnss::multpath_coef(const int& bC, double& cC,
			      const int& b1, double& c1,
                              const int& b2, double& c2)
{
  if( b1 == b2 ){ c1 = c2 = 0.0; return -1; } // phase should be different!
  t_gsys gsys( this->gsys() );

  double pfacC = pow(gsys.factor( bC ), 2);
  double pfac1 = pow(gsys.factor( b1 ), 2);
  double pfac2 = pow(gsys.factor( b2 ), 2);

  double denominator = (pfac1-pfac2)*pfacC;

  cC = 1.0;
  c1 = - (pfac1*(pfacC+pfac2))/denominator;
  c2 = + (pfac2*(pfacC+pfac1))/denominator;

//  cC = 1.0;
//  c1 = - (2*pfacC)/(pfac1 - pfac2) - 1;
//  c2 = + (2*pfacC)/(pfac1 - pfac2);

  return 0;
}


// Ionosphere-linear code combination [m]
// ----------
double t_gobsgnss::P3(const int& b1, const int& b2) // const
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  double coef1, coef2;
   
  ionofree_coef( b1, coef1, b2, coef2 );
     
  double lc = code_lc( b1, coef1, b2, coef2 ); // still dual-frequency only

//  cout << "t_gobsgnss::P3() " << sat() << " " << fixed << setprecision(3) << lc << endl;

  _gmutex.unlock(); return lc;

}


// get Geometry-free phase combination [m]
// ----------
double t_gobsgnss::P4(const int& b1, const int& b2) // const
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  double coef1, coef2;

  geomfree_coef( b1, coef1, b2, coef2 );
  double lc = code_lc( b1, coef1, b2, coef2 ); // still dual-frequency only
  _gmutex.unlock(); return lc;
}

// Ionosphere-free linear phase combination [m] !
// ----------
double t_gobsgnss::L3(const int& b1, const int& b2) // const
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

   if (gsys() == GLO && double_eq(_channel, DEF_CHANNEL)) {
//      cout << epoch().str_hms() << " " << sat() << " " << _channel << endl;      
      _gmutex.unlock(); 
      return 0.0;
   }   
   
  double coef1, coef2;
   
  ionofree_coef( b1, coef1, b2, coef2 );

  double lc = phase_lc( b1, coef1, b2, coef2 ); // still dual-frequency only

//  cout << "t_gobsgnss::L3() " << sat() << " " << fixed << setprecision(3) << lc << endl;

  _gmutex.unlock(); return lc;

}


// get Geometry-free combination [m] !
// ----------
double t_gobsgnss::L4(const int& b1, const int& b2) // const
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
   if (gsys() == GLO && double_eq(_channel, DEF_CHANNEL)) {
      _gmutex.unlock(); 
      return 0.0;
   }
   
  double coef1, coef2;

  geomfree_coef( b1, coef1, b2, coef2 );
  double lc = phase_lc( b1, coef1, b2, coef2 ); // still dual-frequency only

#ifdef DEBUG   
  cout << " L4 = " << fixed << setprecision(3)
       << setw(16)      << lc
       << setw(16)      << phase_lc(b1, 1, b2, 0)  // first  band
       << setw(16)      << phase_lc(b1, 0, b2, 1)  // second band
       << endl;
#endif   

  _gmutex.unlock(); return lc;
}


// Melbourne-Wuebenna combination phase & code [m] !
// ----------
double t_gobsgnss::MW(const int& b1, const int& b2)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

   if (gsys() == GLO && double_eq(_channel, DEF_CHANNEL)) {
      _gmutex.unlock(); 
      return 0.0;
   }
   
  double coefL1, coefL2, coefP1, coefP2;
   
  widelane_coef( b1, coefL1, b2, coefL2 );
  narrlane_coef( b1, coefP1, b2, coefP2 );

  double wideL = phase_lc( b1, coefL1, b2, coefL2 );
  double narrP =  code_lc( b1, coefP1, b2, coefP2 );
  
  if( double_eq(narrP, NULL_GOBS) ){ _gmutex.unlock(); return NULL_GOBS; }
  if( double_eq(wideL, NULL_GOBS) ){ _gmutex.unlock(); return NULL_GOBS; }

//  double lc = wideL - narrP;
  double lc = narrP - wideL;  

#ifdef DEBUG
  cout << " coefL1 = " <<  coefL1
       << " coefL2 = " <<  coefL2
       << " coefP1 = " <<  coefP1
       << " coefP2 = " <<  coefP2
       << " b1 = "     <<  b1
       << " b2 = "     <<  b2     
       << "     lc = " <<  lc     << endl;
#endif

  _gmutex.unlock(); return lc;
}

// get wide-lane combination for phase [m]!
// ---------------------------------------
double t_gobsgnss::LWL(const int& b1, const int& b2)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
   if (gsys() == GLO && double_eq(_channel, DEF_CHANNEL)) {
      _gmutex.unlock(); 
      return 0.0;
   }
   
  double coef1, coef2;

  widelane_coef( b1, coef1, b2, coef2 );
  double lc = phase_lc( b1, coef1, b2, coef2 ); // still dual-frequency only

#ifdef DEBUG   
  cout << " L4 = " << fixed << setprecision(3)
       << setw(16)      << lc
       << setw(16)      << phase_lc(b1, 1, b2, 0)  // first  band
       << setw(16)      << phase_lc(b1, 0, b2, 1)  // second band
       << endl;
#endif   

  _gmutex.unlock(); return lc;   
}

// get wide-lane combination for phase [m]!
// ---------------------------------------
double t_gobsgnss::LNL(const int& b1, const int& b2)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
   if (gsys() == GLO && double_eq(_channel, DEF_CHANNEL)) {
      _gmutex.unlock(); 
      return 0.0;
   }
   
  double coef1, coef2;

  narrlane_coef( b1, coef1, b2, coef2 );
  double lc = phase_lc( b1, coef1, b2, coef2 ); // still dual-frequency only

#ifdef DEBUG   
  cout << " L4 = " << fixed << setprecision(3)
       << setw(16)      << lc
       << setw(16)      << phase_lc(b1, 1, b2, 0)  // first  band
       << setw(16)      << phase_lc(b1, 0, b2, 1)  // second band
       << endl;
#endif   

  _gmutex.unlock(); return lc;   
}

//get mulitpath LC
//---------------------
double t_gobsgnss::MP(const GOBS& ctype, const int& b1, const int& b2)
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  double coef1, coef2, coefC;
   
  double C = obs_C(ctype);    // code obs   
  int bC = t_gobs::gobs2band(ctype); // code band
  if( double_eq(C, NULL_GOBS) ){ _gmutex.unlock(); return NULL_GOBS; }

  multpath_coef( bC, coefC, b1, coef1, b2, coef2 );
   
  double L = phase_lc( b1, coef1, b2, coef2 );  
  if( double_eq(L, NULL_GOBS) ){ _gmutex.unlock(); return NULL_GOBS; }


  double lc = coefC*C + L;

#ifdef DEBUG
  cout << "MP" << bC << t_gobs::gobs2str(ctype)
       << fixed << setprecision(3)
       <<    ": coef1 = " << setw(7) << coef1
       << "     coef2 = " << setw(7) << coef2
       << "         C = " << setw(7) << C
       << "         L = " << setw(7) << L
       << "        lc = " << setw(7) << lc    
       << endl;
#endif

  _gmutex.unlock(); return lc;
}


// valid
// ----------
bool t_gobsgnss::valid()
{
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
  bool tmp = this->valid();
   
  _gmutex.unlock(); return tmp;
   
}


// clean data
// ----------
void t_gobsgnss::clear()
{ 
#ifdef BMUTEX   
  boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();

  this->_clear();
   
  _gmutex.unlock();
}


// clean internal function
// ----------
void t_gobsgnss::_clear()
{
  _staid.clear();
  _epoch = t_gtime::first_time();
}


// clean internal function
// ----------
bool t_gobsgnss::_valid() const
{   
  if( _staid.empty() || 
      _staid == ""   || 
      _epoch == t_gtime::first_time() ) return false;

  return true; 
}

// Band of code/phase
// ------------------------------
void t_gobsgnss::nbands(pair<int,int>& nb)
{
   for (int b=1; b<=8; b++)
   {
      if (cod_id(b) != X) nb.first++;
      if (pha_id(b) != X) nb.second++;      
   }
}

// Operator -
// -----------------------------
t_gobsgnss t_gobsgnss::operator-(t_gobsgnss& gobsgnss)
{
   t_gobsgnss sd;
   double dP, dL;
   
   // Code difference
   for (int b = 0; b<9; b++){
      GOBS type = cod_id(b);
      if ( type == X) continue;
      dP = this->obs_C(1) - gobsgnss.obs_C(b); // PROC JE ZDE natvrdo prvni frenkvence ?
      sd.addobs(type, dP);
   }
   
   // Phase difference
   for (int b = 0; b<9; b++){
      GOBS type =  pha_id(b);
      if ( type == X) continue;
      dL = this->obs_L(1) - gobsgnss.obs_L(b); // PROC JE ZDE natvrdo prvni frekvence ?
      sd.addobs(type, dL);
   }   

   string satSD = this->sat() + gobsgnss.sat();
   sd.sat(satSD);
     
   return sd;
}

// Validity test for GLONASS data
// ---------------
bool t_gobsgnss::_valid_obsL()
{
   if (this->gsys() == GLO && _channel == 255)
     return false;
   else return true;
}

