
/* ----------------------------------------------------------------------
 * G-Nut - GNSS software development library
 * 
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  This file is part of the G-Nut C++ library.
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License as
  published by the Free Software Foundation; either version 3 of
  the License, or (at your option) any later version.
 
  This library is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, see <http://www.gnu.org/licenses>.

-*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <iomanip>
#include <sstream>

#include "gdata/gsatdata.h"
#include "gmodels/gephplan.h"
#include "gutils/gtypeconv.h"
#include "gutils/gsysconv.h"

using namespace std;

// -----------
t_gsatdata::t_gsatdata(string site, string sat, const t_gtime& t)
  : t_gobsgnss(site,sat,t),
    _satcrd(0.0,0.0,0.0),
    _clk(0.0),
    _ele(0.0),
    _azi(0.0),
    _rho(0.0),
    _eclipse(false)
{ 
  id_type(t_gdata::SATDATA);
  id_group(t_gdata::GRP_OBSERV);
  _lastEcl = t_gtime::first_time();
}


// -----------
t_gsatdata::t_gsatdata(const t_gobsgnss& obs)
  : t_gobsgnss(obs), // t_gobsgnss(obs.site() ,obs.sat(), obs.epoch() ),
    _satcrd(0.0,0.0,0.0),
    _clk(0.0),
    _ele(0.0),
    _azi(0.0),
    _rho(0.0),
    _eclipse(false)
{        
  id_type(t_gdata::SATDATA);
  id_group(t_gdata::GRP_OBSERV);
  _lastEcl = t_gtime::first_time();
}


// -----------
t_gsatdata::~t_gsatdata()
{
#ifdef DEBUG
  cout << "GSATDATA - destruct POCAT: " << site() << " " << sat() << " "
       << epoch().str("  %Y-%m-%d %H:%M:%S[%T] ") << fixed << setprecision(3) << endl;
   
  vector<GOBS> v_obs = this->obs();
  vector<GOBS>::iterator itOBS = v_obs.begin();
  for( ; itOBS != v_obs.end(); ++itOBS ) 
      cout << " " << t_gobs::gobs2str( *itOBS ) << ":" << this->getobs( *itOBS );
  cout << endl;

  cout << "GSATDATA - destruct KONEC: " << site() << " " << sat() << " "
       << epoch().str("  %Y-%m-%d %H:%M:%S[%T] ");
  cout.flush();   
#endif   
}


// -----------
void t_gsatdata::addcrd( const t_gtriple& crd )
{
  gtrace("t_gsatdat::addcrd");
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  _satcrd = crd;
  _gmutex.unlock();  
  return;
}

//----------
int t_gsatdata::addprd( t_gallnav* gnav)
{
  gtrace("t_gsatdat::addprd");
   
   t_gsys gsys(this->gsys());
   GSYS gs = gsys.gsys();
   int b1 = gsys.first_band(gs,1);
   int b2 = gsys.second_band(gs,b1);

   double P3 = this->P3(b1, b2);
   double L3 = this->L3(b1, b2);

#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();
   
   //test for observations availability
   if( P3 == 0 || L3 == 0 || gnav == 0 ) {
//      cout << "Add prd " << epoch().str_hms() << " " << sat() << " " << P3 << " " << L3 << " " << gnav << endl;
      _gmutex.unlock();
      return -1;
   }     
   
   double xyz[3]  = {0.0, 0.0, 0.0};
   double vel[3]  = {0.0, 0.0, 0.0};
   double var[3]  = {0.0, 0.0, 0.0};
   double clk =    0.0;
   double dclk =   0.0;
   double clkrms = 0.0;   
   
   string satname( _satid );
   if( satname.substr(0,1) != "G" &&
       satname.substr(0,1) != "R" &&
       satname.substr(0,1) != "E" &&
       satname.substr(0,1) != "J" &&
       satname.substr(0,1) != "S" &&       
       satname.substr(0,1) != "C" )
   {  _gmutex.unlock(); return -1; }

   t_gtime epoT(t_gtime::GPS);
   double satclk = 0.0;
   double satclk2 = 1.0;

   while (fabs(satclk-satclk2) > 1.e-3/CLIGHT){
      satclk2 = satclk;
      epoT = _epoch - P3/CLIGHT - satclk;
      int irc = gnav->clk( satname, epoT, &clk, &clkrms, &dclk );

      if( irc < 0 ){

         if( _log )
	     _log->comment(2, "gppp"," satelite " + satname 
			 + _epoch.str_ymdhms(" clocks not calculated for epoch "));
//	 cout << satname + _epoch.str_ymdhms(" clocks not calculated for epoch ") << endl;
	 _gmutex.unlock(); return -1;
      }
      satclk = clk;

   }

   int irc = gnav->pos( satname, epoT, xyz, var, vel );

   if( irc < 0 ) {
      if( _log )
	  _log->comment(2, "gppp"," satelite " + satname
			 + _epoch.str_ymdhms(" coordinates not calculated for epoch "));
//	 cout << satname + _epoch.str_ymdhms(" coordinates not calculated for epoch ") << endl;
      _gmutex.unlock(); return -1;
   }  

   t_gtriple txyz(xyz);   

   // relativistic correction
   // WARNING: GLONASS clk already include the correction from brd!!!!!
   if( satname.substr(0,1) != "R" ){
     satclk -= 2.0 * ( txyz[0]*vel[0] + txyz[1]*vel[1] + txyz[2]*vel[2] ) /CLIGHT /CLIGHT;
   }
     
   // filling gsatdata
   _satcrd = txyz;
   _clk    = satclk*CLIGHT;

   this->addecl(txyz, epoT);

#ifdef DEBUG      
   cout << satname 
        << " CRD " << fixed << setprecision(3)
        <<  "  "   << epoT.str_ymdhms()
        << " X "   << setw(14) << txyz[0]
        << " Y "   << setw(14) << txyz[1]
        << " Z "   << setw(14) << txyz[2]
        << " T "   << setw(14) << satclk*1000.0*1000.0
        << endl;
//   int ooo; cin >> ooo;
#endif

#ifdef BMUTEX
   lock.unlock();
#endif
   _gmutex.unlock(); return 1;
}

// Compute rho, ele, azi ...
// -----------------------
int t_gsatdata::cmpVal(t_gtriple& xyz)
{
   gtrace("t_gsatdata::cmpVal");
   
   if( double_eq(_satcrd[0], 0.0) ||
       double_eq(_satcrd[0], 0.0) ||
       double_eq(_satcrd[0], 0.0) ) {
   cerr << "Satellite position has not been calculated" << endl;
      return -1;
   }
   
   
   if( double_eq(xyz[0], 0.0) ||
       double_eq(xyz[0], 0.0) ||
       double_eq(xyz[0], 0.0) ) {
   cerr << "Station position has not been calculated" << endl;   
      return -1;
   }
    
   t_gtriple neu;

   xyz2neu(_satcrd, xyz, neu);
      
   // Correct Earth rotation
   ColumnVector xRec(3);
   double rho0 = (_satcrd.crd_cvect() - xyz.crd_cvect()).norm_Frobenius();
   double dPhi = OMEGA * rho0 / CLIGHT;

   xRec(1) = xyz[0] * cos(dPhi) - xyz[1] * sin(dPhi);
   xRec(2) = xyz[1] * cos(dPhi) + xyz[0] * sin(dPhi);
   xRec(3) = xyz[2];  
		 
   double tmp = (_satcrd.crd_cvect() - xRec).norm_Frobenius();
   _rho = tmp;

   double NE2 = neu[0]*neu[0]+neu[1]*neu[1];
   double ele = acos(sqrt(NE2)/_rho);
   if( neu[2]<0.0 ) {
      ele *= - 1.0;
   }

   if( sqrt(NE2)/_rho > 1.0 )
        _ele = 0.0;
   else _ele = ele;

   double azi = atan2(neu[1], neu[0]);
   if (azi < 0) azi += 2*M_PI;
      _azi = azi;
   
   return 1;
}



// -----------
void t_gsatdata::addclk( double clk )
{
  gtrace("t_gsatdat::addclk");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  _clk = clk;
  _gmutex.unlock();
}


// -----------
void t_gsatdata::addele( double ele )
{
  gtrace("t_gsatdat::addele");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  _ele = ele;
  _gmutex.unlock();  
}


// -----------
void t_gsatdata::addazi( double azi )
{
  gtrace("t_gsatdat::addazi");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  _azi = azi;
  _gmutex.unlock();  
}


// -----------
void t_gsatdata::addrho( double rho )
{
  gtrace("t_gsatdat::addrho");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  _rho = rho;
  _gmutex.unlock();  
}


// -----------
t_gtriple t_gsatdata::satcrd()
{
  gtrace("t_gsatdat::satcrd");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  t_gtriple crd(_satcrd[0], _satcrd[1], _satcrd[2] );
  _gmutex.unlock();  

  return crd;
}


// -----------
double t_gsatdata::clk()
{
  gtrace("t_gsatdat::clk");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  double tmp = _clk;
  _gmutex.unlock();  
  return tmp;
}


// -----------
double t_gsatdata::ele()
{
  gtrace("t_gsatdat::ele");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  double tmp = _ele;
  _gmutex.unlock(); 
  return tmp;
}

// -----------
double t_gsatdata::ele_deg()
{
  gtrace("t_gsatdat::ele_deg");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  double tmp = _ele*180.0/M_PI;
  _gmutex.unlock();
  return tmp;
}


// -----------
double t_gsatdata::azi()
{
  gtrace("t_gsatdat::azi");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  double tmp = _azi;
  _gmutex.unlock();
  return tmp;
}


// -----------
double t_gsatdata::rho()
{
  gtrace("t_gsatdat::rho");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  double tmp = _rho;
  _gmutex.unlock();
  return tmp;
}


// valid
// ----------
bool t_gsatdata::valid()
{
  gtrace("t_gsatdat::valid");
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  bool tmp = this->_valid();
  _gmutex.unlock();
  return tmp;
}


// clean data
// ----------
void t_gsatdata::clear()
{ 
  gtrace("t_gsatdat::clear");   
   
#ifdef BMUTEX
   boost::mutex::scoped_lock lock(_mutex);
#endif
  _gmutex.lock();  
  this->_clear();
  _gmutex.unlock();  
}


// clean internal function
// ----------
void t_gsatdata::_clear()
{
  gtrace("t_gsatdat::_clear");   
   
  t_gobsgnss::_clear();
//  _satcrd[0] = 0.0;
//  _satcrd[1] = 0.0;
//  _satcrd[2] = 0.0;
  _ele = 0.0;
  _azi = 0.0;
  _rho = 0.0;
  _clk = 0.0;
}


// clean internal function
// ----------
bool t_gsatdata::_valid() const
{   
  gtrace("t_gsatdat::_valid");   
   
   if( _rho   == 0.0  ||   // single validity identification for gsatdata!
       t_gobsgnss::_valid() ) return false;

  return true; 
}

// get eclipsing
// ---------------------
bool t_gsatdata::ecl()
{
  gtrace("t_gsatdat::ecl");   
   
   return _eclipse;
}

// determi ne wheather eclipsed or not
// -------------------------------------
void t_gsatdata::addecl(t_gtriple& sat, t_gtime& epoch)
{
  gtrace("t_gsatdat::addecl");   
   
   t_gephplan plan;
   double mjd = epoch.dmjd();
   
   t_gtriple sun = plan.sunPos(mjd);
   
   // Unit vector Earth - Sun 
   ColumnVector eSun = sun.unitary();
     
   // Unit vector Earth - Satellite
   ColumnVector eSat = sat.unitary();
   
   double cosin = dotproduct(eSun, eSat);
   if (cosin < 0)
   {
      double r = sat.crd_cvect().norm_Frobenius();
      double val = r*sqrt(1-cosin*cosin);
      if (val < Aell) {
	 _eclipse = true;
	 _lastEcl = epoch;
	 return;
      }else{
	 _eclipse = false;
      }
   }
   
   double tdiff = epoch.diff(_lastEcl);
   if (tdiff <= 1800) _eclipse = true;
}

