
#ifndef GEXTNEWMAT_H
#define GEXTNEWMAT_H

/* ----------------------------------------------------------------------
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  2012-05-23 /GG: created

  Solve: At * p
  Input: design matrix A, weight vector p (diagonal from P)
  Output: result of multiplication B

-*/

inline int solve_atp(const Matrix &A, const ColumnVector &p, Matrix &B)
{
    const int m = A.Nrows();
    const int n = A.Ncols();

    for(int i=1;i<=m;i++)
    {
        for(int j=1;j<=n;j++)
        {
            B(j,i) = A(i,j)*p(i);
        }
    }
    return 1;
}

/*
   Solve: |A1|       |l1|
          |A2| * x - |l2|

   Input: block design matrix A1, A2, vector x.
          block measurement vector l1, l2
   Output: result of multiplication vector e
*/

inline int residual_vec(const Matrix &A1, const Matrix &A2, const ColumnVector &l1,
                 const ColumnVector &l2, const ColumnVector &x, ColumnVector &e)
{
     int ofset = A1.Nrows();
//     int m = A1.Nrows()+A2.Nrows();
//     int n = x.Nrows();

     e=0.0;

     for(int i=1;i<=A1.Nrows();i++)
     {
         for(int k=1;k<=A1.Ncols();k++)
         {
             e(i)      +=A1(i,k)*x(k);
             e(i+ofset)+=A2(i,k)*x(k);
         }
         e(i)       -= l1(i);
         e(i+ofset) -= l2(i);
     }

    return 1;
}

/*
   Solve: |A1|
          |A2| * sxx * |A1t,A2t|

   Input:  block design matrix A1, A2,
           covariance matrix of unknown sxx
   Output: measurement covariance matrix sll
*/

inline int measurement_covmat(const Matrix &A1, const Matrix &A2, const Matrix &sxx, Matrix &sll)
{
    int ofset = A1.Nrows();
    int m = A1.Nrows()+A2.Nrows();
    int n = sxx.Ncols();
    Matrix tmp(m,n);
    tmp = 0.0;
    sll = 0.0;

    for(int i=1;i<=A1.Nrows();i++)
    {
        for(int j=1;j<=sxx.Ncols();j++)
        {
            for(int k=1;k<=A1.Ncols();k++)
            {
                tmp(i,j)      +=A1(i,k)*sxx(k,j);
                tmp(i+ofset,j)+=A2(i,k)*sxx(k,j);
            }
        }
    }

    for(int i=1;i<=tmp.Nrows();i++)
    {
        for(int j=1;j<=A1.Nrows();j++)
        {
            for(int k=1;k<=tmp.Ncols();k++)
            {
                sll(i,j)       += tmp(i,k)*A1(j,k);
                sll(i,j+ofset) += tmp(i,k)*A2(j,k);
            }
        }
    }
    return 1;
}

/*
   Solve: |A1|
          |A2| * sxx * |A1t,A2t|

   Input:  block design matrix A1, A2,
           covariance symmetric matrix of unknown sxx
   Output: measurement covariance symmetric matrix sll
*/

inline int measurement_covmat(const Matrix &A1, const Matrix &A2, const SymmetricMatrix &sxx, SymmetricMatrix &sll)
{
    int ofset = A1.Nrows();
    int m = A1.Nrows()+A2.Nrows();
    int n = sxx.Ncols();

    Matrix tmp(m,n); tmp = 0.0;
    sll=0.0;

    double t1;
    double t2;

    for(int i=1;i<=A1.Nrows();i++)
    {
        for(int j=1;j<=sxx.Ncols();j++)
        {
            for(int k=j;k<=A1.Ncols();k++)
            {
                tmp(i,j)      +=A1(i,k)*sxx(k,j);
                tmp(i+ofset,j)+=A2(i,k)*sxx(k,j);
            }
            for(int k=1;k<j;k++)
            {
                tmp(i,j)      +=A1(i,k)*sxx(j,k);
                tmp(i+ofset,j)+=A2(i,k)*sxx(j,k);
            }
        }
    }

    for(int i=1;i<=tmp.Nrows();i++)
    {
        for(int j=1;j<=A1.Nrows();j++)
        {
            t1=t2=0.0;
            for(int k=1;k<=tmp.Ncols();k++)
            {
                t1 += tmp(i,k)*A1(j,k);
                t2 += tmp(i,k)*A2(j,k);

                if(i>=j)
                {
                    sll(i,j)=t1;
                }
                if(i>=j+ofset)
                {
                    sll(i,j+ofset)=t2;
                }
            }
        }
    }
    return 1;
}

/*
   Solve: |A1|
          |A2| * sxx * |A1t,A2t|

   Input:  block design matrix A1, A2,
           covariance symmetric matrix of unknown sxx
   Output: measurement covariance vector sll
*/

inline int measurement_covmat(const Matrix &A1, const Matrix &A2, const SymmetricMatrix &sxx, ColumnVector &sll)
{
    int ofset = A1.Nrows();
    int m = A1.Nrows()+A2.Nrows();
    int n = sxx.Ncols();

    Matrix tmp(m,n); tmp = 0.0;
    sll=0.0;

    double t1;
    double t2;

    for(int i=1;i<=A1.Nrows();i++)
    {
        for(int j=1;j<=sxx.Ncols();j++)
        {
            for(int k=j;k<=A1.Ncols();k++)
            {
                tmp(i,j)      +=A1(i,k)*sxx(k,j);
                tmp(i+ofset,j)+=A2(i,k)*sxx(k,j);
            }
            for(int k=1;k<j;k++)
            {
                tmp(i,j)      +=A1(i,k)*sxx(j,k);
                tmp(i+ofset,j)+=A2(i,k)*sxx(j,k);
            }
        }
    }

    for(int i=1;i<=tmp.Nrows();i++)
    {
        for(int j=1;j<=A1.Nrows();j++)
        {
            t1=t2=0.0;
            for(int k=1;k<=tmp.Ncols();k++)
            {
                t1 += tmp(i,k)*A1(j,k);
                t2 += tmp(i,k)*A2(j,k);

                if(i==j)
                {
                    sll(i)=t1;
                }
                if(i==j+ofset)
                {
                    sll(i)=t2;
                }
            }
        }
    }
    return 1;
}

#endif // GEXTNEWMAT_H
