
#ifndef GSYS_H
#define GSYS_H
 
/* ----------------------------------------------------------------------
  (c) 2011 Geodetic Observatory Pecny, http://www.pecny.cz (gnss@pecny.cz)
      Research Institute of Geodesy, Topography and Cartography
      Ondrejov 244, 251 65, Czech Republic

  Purpose: definition of GNSS constellations and observation frequencies
  Version: $ Rev: $

  2012-09-26 /JD: created

-*/

#include <string>
#ifdef BMUTEX
#include <boost/thread/mutex.hpp>
#endif

#include "gutils/gconst.h"
#include "gutils/gmutex.h"

using namespace std;

// GNSS base frequencies [Hz]
// --------------------------
#define  GPS_FRQ      10230000.0       // GPS     base frequency
#define  GLO_FRQ     178000000.0       // GLONASS base frequency
#define  GAL_FRQ      10230000.0       // GALILEO base frequency
#define  BDS_FRQ      10230000.0       // BEIDOU  base frequency
#define  QZSS_FRQ     10230000.0       // QZSS    base frequency
#define  SBAS_FRQ     10230000.0       // SBAS    base frequency

// GNSS frequency multiplier factors (dimensionless)
// -------------------------------------------------
#define  G01_MLT          154.0
#define  G02_MLT          120.0
#define  G05_MLT          115.0
#define  R01_MLT            9.0
#define  R02_MLT            7.0
#define  R01_STEP      562500.0
#define  R02_STEP      437500.0
#define  E01_MLT          154.0
#define  E05_MLT          115.0
#define  E06_MLT          125.0
#define  E07_MLT          118.0
#define  E08_MLT          116.5
#define  C01_MLT          152.6  
#define  C06_MLT          124.0
#define  C07_MLT          118.0
#define  J01_MLT          154.0
#define  J02_MLT          120.0
#define  J05_MLT          115.0
#define  J06_MLT          125.0
#define  S01_MLT          154.0
#define  S05_MLT          115.0

// GNSS derived frequencies [Hz]
// -----------------------------
#define  G01_F      G01_MLT * GPS_FRQ                //  L1    - GPS NAVSTAR
#define  G02_F      G02_MLT * GPS_FRQ                //  L2    - GPS NAVSTAR
#define  G05_F      G05_MLT * GPS_FRQ                //  L5    - GPS NAVSTAR

#define  R01_F(a)   R01_MLT * GLO_FRQ + R01_STEP*(a) //  G1    - GLONASS base frequency
#define  R02_F(a)   R02_MLT * GLO_FRQ + R02_STEP*(a) //  G2    - GLONASS base frequency
#define  R03_F      1202250000                       //  G3    - GLONASS CDMA frequency

#define  E01_F      E01_MLT * GAL_FRQ                //  E1    - Galileo
#define  E05_F      E05_MLT * GAL_FRQ                //  E5a   - Galileo
#define  E06_F      E06_MLT * GAL_FRQ                //  E6    - Galileo
#define  E07_F      E07_MLT * GAL_FRQ                //  E5b   - Galileo
#define  E08_F      E08_MLT * GAL_FRQ                //  E5a+b - Galileo

#define  C01_F      C01_MLT * BDS_FRQ                //  E1    - BeiDou
#define  C06_F      C06_MLT * BDS_FRQ                //  E6    - BeiDou
#define  C07_F      C07_MLT * BDS_FRQ                //  E5b   - BeiDou

#define  J01_F      J01_MLT * QZSS_FRQ               //  L1    - QZSS
#define  J02_F      J02_MLT * QZSS_FRQ               //  L2    - QZSS
#define  J05_F      J05_MLT * QZSS_FRQ               //  L5    - QZSS
#define  J06_F      J06_MLT * QZSS_FRQ               //  LEX   - QZSS

#define  S01_F      S01_MLT * SBAS_FRQ               //  L1    - SBAS
#define  S05_F      S05_MLT * SBAS_FRQ               //  L5    - SBAS

// OBSOLETE INFORMATIONS -- NOT YET !?
#define GPSLAMB1    CLIGHT/(G01_F)     // GPS L1 wave length [m]
#define GPSLAMB2    CLIGHT/(G02_F)     // GPS L2 wave length [m]
#define GPSLAMB3   ( ((G01_F*G01_F) / (G01_F*G01_F - G02_F*G02_F)) * GPSLAMB1 - ((G02_F*G02_F) / (G01_F*G01_F - G02_F*G02_F)) * GPSLAMB2 )  // GPS ionospher-free wave lenght [m]
#define LAMBWL      CLIGHT/(G01_F-G02_F)   // GPS wide-lane wave length [m]


// GNSS systems and augmentations
// ------------------------------
enum GSYS { // GXX = -1,
            GPS, GAL, GLO, BDS, QZSS, SBAS, GNSS };

// GNSS frequencies
// -----------------
enum GFRQ { // FXX = -1,
            G01, G02, G05,            // GPS
            R01, R02,                 // GLONASS
            E01, E05, E07, E08, E06,  // Galileo
            C01,      C07, C06,       // BeiDou
            J01, J02, J05, J06,       // QZSS
            S01, S05,                 // SBAS
            LAST_GFRQ
          };

// class 
// ---------- 
class t_gsys {
   
 public:
   t_gsys(GSYS sys);
   t_gsys(string sys);
   t_gsys(char c);
   t_gsys(){};
  ~t_gsys(){};

   // general (static) functions
   static string gsys2str(GSYS sys);             // convert GSYS enum to GSYS string
   static char gsys2char(GSYS sys);              // convert GSYS enum to GSYS char
   static GSYS str2gsys(string s);               // convert GSYS string to GSYS enum
   static char str2char(string s);               // convert GSYS string to GSYS char
   static string char2str(char c);               // convert GSYS char to GSYS string
   static GSYS char2gsys(char c);                // convert GSYS char to GSYS enum   
   
   // get system bands
   static int first_band(GSYS gs, int b);        // get second band to the first one
   static int second_band(GSYS gs, int i);       // get second band to the first one
   static int three_band(GSYS gs, int& b1, int& b2, int& b3); // get three bands number w.r.t. GSYS

   // convert satellite name
   static string eval_sat(string sat, GSYS sys = GPS);  // convert full sat string

// static double frequency(  int i, string sat); // get sat-specific frequency  for band i
// static double wavelength( int i, string sat); // get sat-specific wavelength for band i
   
   double factor( int i );                       // get system-specific factor     for band i
   double frequency( int i );                    // get system-specific frequency  for band i
   double wavelength( int i );                   // get system-specific wavelength for band i

   void   from_string( string sys );             // set GSYS from string
   void   from_gsys( GSYS     sys );             // set GSYS from enum
 
   GSYS   gsys()const{return _gsys;}             // get GSYS enum
   string  str()const{return gsys2str(_gsys);}   // get GSYS string
   
   bool operator==(const string& sys) const;     // overloaded equivalence operator
   bool operator==(const GSYS&   sys) const;     // overloaded equivalence operator
   
 protected:
// GSYS   _to_gsys( string sys );                // internal conversion function
// string _to_string( GSYS sys );                // internal conversion function
   
 private:
   GSYS            _gsys;
   t_gmutex        _gmutex;
#ifdef BMUTEX   
   boost::mutex    _mutex;
#endif
};


// class
// ----------
class t_gfreq {

 public:
   t_gfreq(){};
  ~t_gfreq(){};

   static GFRQ   str2gfreq( string frq );      // get GFRQ enum from string
   static string gfreq2str( GFRQ frq );        // get string from GFRQ enum
   
//   static lambda_L1( GSYS )

 private:
   GFRQ    _gfreq;
//   static boost::mutex   _mutex;

};


#endif // GFREQ_H